/******************************************************************/
/* This module handles different disc image formats that are used */
/* to store data from Amstrad discs                               */
/******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "diskimg.h"
#include "dsk.h"
#include "extdsk.h"
#include "filexfer.h"

#define DEBUG

static char     Header[256];

int     DiskImage_OpenDiskImage(void **pDriveData,char *DiskImageName)
{
FILE *fh;
int	DiskType;

	DSK_UNIT_DATA	*pUnit;

	pUnit = (DSK_UNIT_DATA *)malloc(sizeof(DSK_UNIT_DATA));

    /* Open Disk image and identify the type */

    if ((fh=fopen(DiskImageName,"rb"))==NULL)
            return FALSE;

    /* read header into memory */

    fseek(fh,0,SEEK_SET);
    fread(&Header[0],256,1,fh);
    fclose(fh);

    if (memcmp(&Header[0],"MV - CPC",6)==0)
    {
            #ifdef DEBUG
            printf("Standard disk image found.\n");
            #endif
            DiskType = STANDARD_DISK_IMAGE;
            goto recognised;
    }

    if (memcmp(&Header[0],"EXTENDED",8)==0)
    {
            #ifdef DEBUG
            printf("Extended disk image found.\n");
            #endif
            DiskType = EXTENDED_DISK_IMAGE;
            goto recognised;
    }

    #ifdef DEBUG
    printf("Failed to recognise disk image format.\n");
    #endif
    return FALSE;

recognised:
		pUnit->DiskImgType = DiskType;

		*pDriveData = pUnit;

        switch (pUnit->DiskImgType)
        {
                case STANDARD_DISK_IMAGE:
                {
                        return DskOpenDiskImage(pUnit,DiskImageName);
                        break;
                }
                case EXTENDED_DISK_IMAGE:
                {
                        return ExtOpenDiskImage(pUnit,DiskImageName);
                        break;
                }
        }


		return FALSE;
}

int     DiskImage_GetSectorDsk(void *pDriveData,SECTOR_POS *pSectorPos, char *pData, int Length)
{
	DSK_UNIT_DATA	*pUnit;

	pUnit = (DSK_UNIT_DATA *)pDriveData;

        switch (pUnit->DiskImgType)
        {
                case STANDARD_DISK_IMAGE:
                {
                        return DskGetSectorDsk(pUnit,pSectorPos,pData, Length);
                        break;
                }
                case EXTENDED_DISK_IMAGE:
		{
                        return ExtGetSectorDsk(pUnit,pSectorPos,pData, Length);
                        break;
                }
        }
        return FALSE;

}

int     DiskImage_PutSectorDsk(void *pDriveData,SECTOR_POS *pSectorPos, char *pData, int Length)
{

	DSK_UNIT_DATA	*pUnit;

	pUnit = (DSK_UNIT_DATA *)pDriveData;

        switch (pUnit->DiskImgType)
        {
                case STANDARD_DISK_IMAGE:
                {
                        return DskPutSectorDsk(pUnit,pSectorPos,pData, Length);
                        break;
                }
                case EXTENDED_DISK_IMAGE:
                {
                        return ExtPutSectorDsk(pUnit,pSectorPos,pData, Length);
                        break;
                }
        }
        return FALSE;

}

int     DiskImage_GetFormatDsk(void *pData,FORMAT_DATA *pFormatData)
{
	DSK_UNIT_DATA	*pUnit;

	pUnit = (DSK_UNIT_DATA *)pData;

        switch (pUnit->DiskImgType)
        {
                case STANDARD_DISK_IMAGE:
                {
                        return DskGetFormatDsk(pUnit,pFormatData);
                        break;
                }
                case EXTENDED_DISK_IMAGE:
                {
                        return ExtGetFormatDsk(pUnit,pFormatData);
                        break;
                }
        }
        return FALSE;
}

void	DiskImage_CloseDiskImage(void **pDriveData)
{
	DSK_UNIT_DATA	*pUnit;

	pUnit = (DSK_UNIT_DATA *)*pDriveData;

	if (pUnit!=NULL)
	{
		/* free disk image filehandle */

		if (pUnit->DiskImageFileHandle!=NULL)
			fclose(pUnit->DiskImageFileHandle);

		/* free unit data */
		free(pUnit);
	}
}
