/***********************************************************************/
/* This module handles reading and writing of data using the standard  */
/* disc image type                                                     */
/***********************************************************************/


/* Code specific to disk images */


/* Standard disk image code */

#include "dsk.h"
#include "dskstruc.h"
#include "filexfer.h"

#define min(a,b) (a<b) ? (a) : (b)

static void	GetTrackHeader(DSK_UNIT_DATA *,int,int);
static int	GetSectorPos(DSKTRACKHEADER *,int);
static int	DskGetSectorSize(int);


static int	DskSectorSizeTable[7]={128,256,512,1024,2048,4096,8192};

int	DskOpenDiskImage(DSK_UNIT_DATA *pDskUnit,char *DskFilename)
{
DSKHEADER	*pDskHeader;
//char		*pDiskImage;
//int			fno;
//int			filesize;

	/* attempt to open disk image */

	if ((pDskUnit->DiskImageFileHandle=fopen(DskFilename,"rb+"))==NULL)
		return FALSE;

	/* read header into memory */

	fseek(pDskUnit->DiskImageFileHandle,0,SEEK_SET);
	fread(&pDskUnit->DskHeader[0],sizeof(DSKHEADER),1,pDskUnit->DiskImageFileHandle);

	/* identify if disk image is standard or extended */


	pDskHeader = (DSKHEADER *)&pDskUnit->DskHeader[0];



	/* Setup data for disk image format found */

	pDskUnit->TrackSize = (pDskHeader->TrackSizeHigh<<8)+pDskHeader->TrackSizeLow;
	pDskUnit->NumSides = pDskHeader->NumSides;


	return TRUE;
}

static void	GetTrackHeader(DSK_UNIT_DATA *pDskUnit,int Track,int Side)
{

	pDskUnit->TrackOffset=(Track*pDskUnit->TrackSize)*pDskUnit->NumSides+sizeof(DSKHEADER);

	if ((pDskUnit->NumSides==2) && (Side==1))
	{
		pDskUnit->TrackOffset+=pDskUnit->TrackSize;
	}

	fseek(pDskUnit->DiskImageFileHandle,(long)pDskUnit->TrackOffset,SEEK_SET);
	fread(&pDskUnit->DskTrackHeader[0],sizeof(DSKTRACKHEADER),1,pDskUnit->DiskImageFileHandle);
}

/* Returns position 0->15 or -1 if error */

static int	GetSectorPos(DSKTRACKHEADER *pDskTrackHeader,int R)
{
int     i;
int	Position=0;

	for (i=0; i<pDskTrackHeader->SPT; i++)
	{
		if (pDskTrackHeader->SectorIDs[i].R==R)
			return Position;

		Position++;
	}
	return -1;
}

/* Return sector size from N value in sector ID */

static int	DskGetSectorSize(int N)
{
	if (N<8)
		return DskSectorSizeTable[N];
	else
		return 0x0200;
}


int	DskGetSectorDsk(DSK_UNIT_DATA *pDskUnit,SECTOR_POS *pSectorPos, char *pData, int Length)
{
int	SectorPos;
int	SectorSize;
DSKTRACKHEADER *pDskTrackHeader;
int CopyLength;

	pDskTrackHeader = (DSKTRACKHEADER *)&pDskUnit->DskTrackHeader[0];

	/* Check sector exists */

	GetTrackHeader(pDskUnit,pSectorPos->Track,pSectorPos->Side);

	if ((SectorPos = GetSectorPos(pDskTrackHeader,pSectorPos->R))==-1)
	{
		#ifdef DEBUG
		printf("Sector not found!\n");
		#endif

		return FALSE;
	}
	SectorSize = DskGetSectorSize(pDskTrackHeader->BPS);

	pDskUnit->SectorOffset = pDskUnit->TrackOffset + sizeof(DSKTRACKHEADER) + (SectorSize*SectorPos);
    if (Length!=-1)
    {
        CopyLength = min(SectorSize, Length);
    }
    else
    {
        CopyLength = SectorSize;
    }

	fseek(pDskUnit->DiskImageFileHandle,(long)pDskUnit->SectorOffset,SEEK_SET);
	fread(pData,CopyLength,1,pDskUnit->DiskImageFileHandle);

	return TRUE;
}

int	DskPutSectorDsk(DSK_UNIT_DATA *pDskUnit,SECTOR_POS *pSectorPos, char *pData, int Length)
{
int	SectorPos;
int	SectorSize;
DSKTRACKHEADER *pDskTrackHeader;
    int CopyLength;

	pDskTrackHeader = (DSKTRACKHEADER *)&pDskUnit->DskTrackHeader[0];

	/* Check sector exists */

	GetTrackHeader(pDskUnit,pSectorPos->Track,pSectorPos->Side);

	if ((SectorPos = GetSectorPos(pDskTrackHeader,pSectorPos->R))==-1)
	{
		#ifdef DEBUG
		printf("Sector not found!\n");
		#endif

		return FALSE;
	}

	SectorSize = DskGetSectorSize(pDskTrackHeader->BPS);
  if (Length!=-1)
    {
        CopyLength = min(SectorSize, Length);
    }
    else
    {
        CopyLength = SectorSize;
    }

	pDskUnit->SectorOffset = pDskUnit->TrackOffset + sizeof(DSKTRACKHEADER) + (SectorSize*SectorPos);

	fseek(pDskUnit->DiskImageFileHandle,(long)pDskUnit->SectorOffset,SEEK_SET);
	fwrite(pData,CopyLength,1,pDskUnit->DiskImageFileHandle);

	return TRUE;
}

int	DskGetFormatDsk(DSK_UNIT_DATA *pDskUnit,FORMAT_DATA *pFormatData)
{
DSKTRACKHEADER *pDskTrackHeader;
DSKHEADER	*pDskHeader;

	/* Simple based on track 0 */
	/* More complex required to allow multi-side discs,
	protected discs with Directories etc */

	GetTrackHeader(pDskUnit,0,0);
	pDskHeader = (DSKHEADER *)&pDskUnit->DskHeader[0];
	pDskTrackHeader = (DSKTRACKHEADER *)&pDskUnit->DskTrackHeader[0];

	pFormatData->FirstSectorID = pDskTrackHeader->SectorIDs[0].R & 0x0ff;
pFormatData->SectorNSize = pDskTrackHeader->BPS & 0x0ff;
	pFormatData->NumberOfSides = pDskHeader->NumSides;
pFormatData->SectorsPerTrack = pDskTrackHeader->SPT;


	return TRUE;
}
