/*****************************************************************************/
/* This module is the top level. It handles reading and writing files in the */
/* following ways:                                                           */
/* It handles, real disc, disk image or files. And to and from CPC filesystem */
/*****************************************************************************/

/* Code to extract/add files to disc images or real disk */
/* change "disk image and real disc" to have types of "CPC FileSystem" */
/* (c) Kevin Thacker */

#define DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "global.h"
#include "filexfer.h"
#include "diskimg.h"
#include "filen.h"
//#include "amiga.h"
#include "host.h"
#include "ams.h"

#include "namelist.h"
#ifdef _WIN32
#include "direct.h"
#endif

#include "wildcard.h"

/* Reading information */
static DRIVE_DATA       DriveData[2];


void    FileXfer_FreeDrive(int DriveNum)
{
DRIVE_DATA *pDriveData;

                pDriveData = &DriveData[DriveNum];

           switch (pDriveData->DriveType)
           {
                 case DISK_IMAGE:
                {
                        DiskImage_CloseDiskImage(&pDriveData->pDriveData);
                        break;
                }
                case REAL_DISK:
                {
                        /*   */
                        break;
          		 }
        }

        if ((pDriveData->DriveType==DISK_IMAGE) || (pDriveData->DriveType==REAL_DISK))
			{
				Ams_FreeDrive(pDriveData);
			}

}

void     FileXfer_InitialiseDrive(int DriveNum,int DriveType,void *pData)
{
	DRIVE_DATA *pDriveData;

	pDriveData = &DriveData[DriveNum];

        pDriveData->DriveType = DriveType;

        switch (DriveType)
        {
                case    REAL_DISK:
                {
                        printf("Real Disk selected.\nReal Disk not supported!\n");
                        break;
                }
                case    DISK_IMAGE:
                {
                        printf("Disk Image selected\n");
                        DiskImage_OpenDiskImage(&pDriveData->pDriveData,(char *)pData);
                        break;
                }
                case    HOST_FILESYSTEM:
                {
#ifdef WIN32			
                        _chdir((char *)pData);
#else
			chdir((char *)pData);
#endif
			printf("Host filesystem selected\n");
                        break;
                }
        }


        if ((DriveType==DISK_IMAGE) || (DriveType==REAL_DISK))
		{
				Ams_InitialiseDrive(pDriveData);
        }
}


/* Opens a file for reading or writing on a specific drive */

CPC_FILEHANDLE *FileXfer_OpenFile(int DriveNum,char *Filename, int Access)
{
	CPC_FILEHANDLE	*pCPCFileHandle;

	pCPCFileHandle = (CPC_FILEHANDLE *)malloc(sizeof(CPC_FILEHANDLE));

	if (pCPCFileHandle!=NULL)
	{
		/* setup file-handle */
		memset(pCPCFileHandle,0,sizeof(CPC_FILEHANDLE));
		pCPCFileHandle->Filename = malloc(strlen(Filename)+1);

		if (pCPCFileHandle->Filename!=NULL)
		{
			strcpy(pCPCFileHandle->Filename, Filename);
		}

		pCPCFileHandle->Access = Access;

		pCPCFileHandle->pDriveData = &DriveData[DriveNum];

        switch (pCPCFileHandle->pDriveData->DriveType)
        {
            case DISK_IMAGE:
            {

                Ams_OpenFile(pCPCFileHandle,Filename,Access);
                        break;
                }
                case REAL_DISK:
                {
                    break;
		        }
	            case HOST_FILESYSTEM:
					{
						OpenFileHost(pCPCFileHandle,Filename,Access);
					}
					break;

		}

	}

	return pCPCFileHandle;
}

int		FileXfer_GetFileSize(CPC_FILEHANDLE *pCPCFileHandle)
{
	switch (pCPCFileHandle->pDriveData->DriveType)
	{
		case	DISK_IMAGE:
		case	REAL_DISK:
			return	Ams_GetFileSize(pCPCFileHandle);

		case	HOST_FILESYSTEM:
			return	GetFileSizeHost(pCPCFileHandle);

        default:
            break;
	}

    return FALSE;
}

int		FileXfer_GetChar(CPC_FILEHANDLE *pCPCFileHandle)
{
        switch (pCPCFileHandle->pDriveData->DriveType)
        {
            case DISK_IMAGE:
               return Ams_GetChar(pCPCFileHandle);

            case REAL_DISK:
            {
/*                                      return GetCharAms(pCPCFileHandle);*/
                                        break;
            }
            break;

            case HOST_FILESYSTEM:
                return GetCharHost(pCPCFileHandle);

            default:
                break;
        }

    return 0;
}

void		FileXfer_PutChar(CPC_FILEHANDLE *pCPCFileHandle,int Ch)
{
        switch (pCPCFileHandle->pDriveData->DriveType)
        {
            case DISK_IMAGE:
            {

                                Ams_PutChar(pCPCFileHandle,Ch);
                                        break;
                                }
                                case REAL_DISK:
                                {
/*                                      PutCharAms(pCPCFileHandle,Ch);*/
                                        break;
                                }
                                case HOST_FILESYSTEM:
                                        PutCharHost(pCPCFileHandle,Ch);
										break;

		}



}


/* Write a sector to the media */
static int     WriteSector(DRIVE_DATA *pDriveData,SECTOR_POS *pSectorPos, char *pData, int Length)
{
#ifdef DEBUG
 //   printf("Writing sector: Drive %d Physical Track %d Physical Side %x (C: %x, H: %x, R: %x, N: %x) \r\n",pSectorPos->Drive,pSectorPos->Track,pSectorPos->Sector,pSectorPos->Side);
#endif

	switch (pDriveData->DriveType)
	{
		case DISK_IMAGE:
		{

			return DiskImage_PutSectorDsk(pDriveData->pDriveData,pSectorPos,pData, Length);
		}
		break;

		case REAL_DISK:
		{
	/*     return PutSectorDisk(pSectorPos->Track,pSectorPos->Sector,pSectorPos->Side,pData);*/
		}
		break;

		case HOST_FILESYSTEM:
		{
			return FALSE;
		}
		break;
	}
    return TRUE;
}


/* Read a sector of data from media */
static int     ReadSector(DRIVE_DATA *pDriveData, SECTOR_POS *pSectorPos, char *pData, int Length)
{
#ifdef DEBUG
//	printf("AMS: Reading sector: Drive %d Track %d Sector %x Side %d\r\n",pSectorPos->Drive,pSectorPos->Track,pSectorPos->Sector,pSectorPos->Side);
#endif

	switch (pDriveData->DriveType)
	{
		case DISK_IMAGE:
		{
			return DiskImage_GetSectorDsk(pDriveData->pDriveData,pSectorPos,pData, Length);
		}
		break;
		case REAL_DISK:
		{
/*            return GetSectorDisk(pSectorPos->Track,pSectorPos->Sector,pSectorPos->Side,pData);*/
		}
		break;
		case HOST_FILESYSTEM:
		{
			return FALSE;
		}
		break;
	}
	return TRUE;
}

BOOL    FileXFer_ReadSector(int DriveNum, SECTOR_POS *pSectorPos, char *pData, int Length)
{
  	DRIVE_DATA *pDriveData;

	pDriveData = &DriveData[DriveNum];

        switch (pDriveData->DriveType)
        {
                case    DISK_IMAGE:
                {
						return ReadSector(pDriveData, pSectorPos, pData, Length);
                        break;
                }
        }

		return FALSE;

}


BOOL    FileXFer_WriteSector(int DriveNum, SECTOR_POS *pSectorPos, char *pData, int Length)
{
  	DRIVE_DATA *pDriveData;

	pDriveData = &DriveData[DriveNum];

        switch (pDriveData->DriveType)
        {
                case    DISK_IMAGE:
                {
						return WriteSector(pDriveData, pSectorPos, pData, Length);
                        break;
                }
        }

		return FALSE;

}

/* Closes a file for reading or writing on a specific drive */

void		FileXfer_CloseFile(CPC_FILEHANDLE *pCPCFileHandle)
{
	if (pCPCFileHandle!=NULL)
	{
        switch (pCPCFileHandle->pDriveData->DriveType)
        {
            case DISK_IMAGE:
            {

               Ams_CloseFile(pCPCFileHandle);
            break;
           }
            case REAL_DISK:
            {
/*                                      CloseFileAms(pCPCFileHandle);*/
                    break;
            }
            case HOST_FILESYSTEM:
								CloseFileHost(pCPCFileHandle);
								break;
                    }

		free(pCPCFileHandle);
	}

}


BOOL	FileXfer_FileFit(int DriveNum, int FileSize)
{
	DRIVE_DATA *pDriveData;

	pDriveData = &DriveData[DriveNum];

        switch (pDriveData->DriveType)
        {
                case    DISK_IMAGE:
                case    REAL_DISK:
                {
						return Ams_FileFit(pDriveData, FileSize);
                        break;
                }
                case    HOST_FILESYSTEM:
                {
						return Host_FileFit(DriveNum, FileSize);
						break;
                }
        }

		return FALSE;
}



/*============================================================================*/

/* generate a list of files matching Wildcard string on the drive specified */
void	FileXfer_GenerateFileList(int Drive, char *WildCard, NAME_LIST_HEADER *pNameList)
{
	DRIVE_DATA *pDriveData;

   pDriveData = &DriveData[Drive];

    switch (pDriveData->DriveType)
    {
        case    REAL_DISK:
        {
                printf("Real Disk selected.\nReal Disk not supported!\n");
                break;
        }
        case    DISK_IMAGE:
        {
                printf("Disk Image selected\n");
                Ams_GenerateFileList(pDriveData->pFSData, WildCard,pNameList);
                break;
        }
        case    HOST_FILESYSTEM:
        {
				printf("Host filesystem selected\n");
                Host_GenerateFileList(WildCard,pNameList);
                break;
        }
    }
}

/*============================================================================*/

void	FileXfer_DisplayDirectory(int Drive,char *FileSpec)
{





}

void	FileXfer_RenameFile(int SourceDrive, char *SourceName, char *DestName)
{
	DRIVE_DATA *pDriveData;

   pDriveData = &DriveData[SourceDrive];

    switch (pDriveData->DriveType)
    {
        case    REAL_DISK:
        {
                printf("Real Disk selected.\nReal Disk not supported!\n");
                break;
        }
        case    DISK_IMAGE:
        {
                printf("Disk Image selected\n");
                Ams_RenameFile(pDriveData->pFSData, 0, SourceName, DestName);
                break;
        }
        case    HOST_FILESYSTEM:
        {
				printf("Host filesystem selected\n");
                //Host_RenameFile();
                break;
        }
    }
}

void	FileXfer_DeleteFile(int SourceDrive, char *Filename)
{
	DRIVE_DATA *pDriveData;

   pDriveData = &DriveData[SourceDrive];

    switch (pDriveData->DriveType)
    {
        case    REAL_DISK:
        {
                printf("Real Disk selected.\nReal Disk not supported!\n");
                break;
        }
        case    DISK_IMAGE:
        {
                printf("Disk Image selected\n");
                Ams_DeleteFile(pDriveData->pFSData, 0, Filename);
                break;
        }
        case    HOST_FILESYSTEM:
        {
				printf("Host filesystem selected\n");
                //Host_RenameFile();
                break;
        }
    }
}


/* convert a filename between filesystems */
void	ConvertFilename(int SourceDrive, int DestDrive, char *pSourceFilename, char *pDestFilename)
{
	switch (DriveData[SourceDrive].DriveType)
	{
		case DISK_IMAGE:
		{
			switch (DriveData[DestDrive].DriveType)
			{
				case DISK_IMAGE:
				{
					/* source and dest are disk images */

					memcpy(pDestFilename, pSourceFilename, strlen(pSourceFilename)+1);
				}
				break;

				case HOST_FILESYSTEM:
				{
					/* source is disk image, dest is host filesystem */
					CPCFilenameToHostFilename(pSourceFilename, pDestFilename);
				}
				break;
			}
		}
		break;

		case HOST_FILESYSTEM:
		{
			switch (DriveData[DestDrive].DriveType)
			{
				case DISK_IMAGE:
				{
					/* source is host filesystem, dest is disk image */
					HostFilenameToCPCFilename(pSourceFilename, pDestFilename);
				}
				break;

				case HOST_FILESYSTEM:
				{
					/* source and dest are host filesystem */
					memcpy(pDestFilename, pSourceFilename, strlen(pSourceFilename)+1);
				}
				break;
			}
		}
		break;
	}
}


/*============================================================================*/


void	FileXfer_CopySingle(int SourceDrive, int DestDrive, char *pSourceFilename)
{
	CPC_FILEHANDLE	*pSource;
	CPC_FILEHANDLE	*pDest;
	int				SourceFileSize;
	int				i;

	char	DestFilename[64];

	/* convert source filename into destination filename */
	ConvertFilename(SourceDrive, DestDrive, pSourceFilename, DestFilename);

	/* open source file for reading */
	pSource = FileXfer_OpenFile(SourceDrive,pSourceFilename,CPC_FILEHANDLE_READ);

	if (pSource!=NULL)
	{
		/* source file opened */

		/* get size of CPC file */
		SourceFileSize = FileXfer_GetFileSize(pSource);

		/* will source file fit onto destination ?? */
		if (FileXfer_FileFit(DestDrive,SourceFileSize))
		{
			/* file will fit on disc */

			/* open output file for writing */
			pDest = FileXfer_OpenFile(DestDrive,DestFilename,CPC_FILEHANDLE_WRITE);

			if (pDest!=NULL)
			{
				/* copy data accross */
				for (i=0; i<SourceFileSize; i++)
				{
					int	ch;

					/* read from cpc */
					ch = FileXfer_GetChar(pSource);

					/* write to host */
					FileXfer_PutChar(pDest,ch);
				}

				/* close output file */
				FileXfer_CloseFile(pDest);
			}
		}
		else
		{
#ifdef DEBUG__
			printf("File will not fit on media!!!\r\n");
#endif
		}

		/* close input file */
		FileXfer_CloseFile(pSource);
	}
}

/*============================================================================*/

/* copy multiple files between source and dest */
void	FileXfer_CopyMultiple(int SourceDrive, int DestDrive, char *WildCard)
{
	NAME_LIST_HEADER *pNameList;
	NAME_LIST_ENTRY	*pEntry;
	char	ParsedWildCard[16];

	/* convert wildcard */
	WildCard_Parse(WildCard, ParsedWildCard);

	/* create name list */
	pNameList = NameList_Create();

	if (pNameList!=NULL)
	{
		/* name list created */

		/* generate list of files on source drive that match wildcard */
		FileXfer_GenerateFileList(SourceDrive,ParsedWildCard,pNameList);

		/* go through list and do each file */
		pEntry = pNameList->pFirstEntry;

		if (pEntry!=NULL)
		{
			do
			{
				/* if filename exists, do it */
				if (pEntry->Name!=NULL)
				{
					FileXfer_CopySingle(SourceDrive, DestDrive, pEntry->Name);
				}

				/* next entry */
				pEntry = pEntry->pNextEntry;
			}
			while (pEntry!=NULL);
		}

		/* delete name list */
		NameList_Delete(pNameList);
	}
}

/*============================================================================*/


/*============================================================================*/

#include "formdsk.h"

extern XDPB DataFormat;

/*int main(int argc, char *argv[])
{


	//CLI_Go();

	{
	char DriveSpec[] = "c:\\windows\\desktop\\stuff";
	char DiskImage[] = "c:\\windows\\desktop\\testz80.dsk";

	//char DriveSpec[] = "c:\\cpcfiles\\demo5";
	//char DiskImage[] = "c:\\windows\\desktop\\present.dsk";

	// format a dsk to data format
	FormatDsk(&DataFormat,DiskImage);

	// initialise drives
	FileXfer_InitialiseDrive(0,HOST_FILESYSTEM,(void *)DriveSpec);
	FileXfer_InitialiseDrive(1,DISK_IMAGE,(void *)DiskImage);

	// copy files
	//FileXfer_CopyMultiple(0,1,"present*.*");
	//FileXfer_CopyMultiple(0,1,"data.*");

	FileXfer_CopyMultiple(0,1,"z80test*.*");

	//FileXfer_RenameFile(1,"DREAMEND.","DREAM.");

//	FileXfer_DeleteFile(1,"DREAM.");

//	FileXfer_DisplayDirectory(1,"*.*");

	FileXfer_FreeDrive(0);
	FileXfer_FreeDrive(1);
	}
	return TRUE;
}

*/
// wildcard match doesn't work on host.
// generates wrong host filenames

int	FileXfer_GetSizeOfNamedFile(int Drive, char *pFilename)
{
	int FileSize = 0;
	CPC_FILEHANDLE *pTempHandle;

	/* open file */
	pTempHandle = FileXfer_OpenFile(Drive, pFilename, CPC_FILEHANDLE_READ);

	if (pTempHandle!=NULL)
	{
		/* get file size */
		FileSize = FileXfer_GetFileSize(pTempHandle);

		/* close file */
		FileXfer_CloseFile(pTempHandle);
	}

	return FileSize;
}
