#include "formdsk.h"

#include <memory.h>
#include <stdio.h>
#include <stdlib.h>


#include "globdef.h"
#include "amsstruc.h"
#include "dskstruc.h"

#define FILLER_BYTE	0x0e5

static char	HeaderText[]="MV - CPCEMU Disk-File\r\nDisk-Info\r\n\0";
static char	TrackText[]="Track-Info\r\n\0";
static int	DskSectorSizeTable[7]={128,256,512,1024,2048,4096,8192};

/* Return sector size from N value in sector ID */

static int	DskGetSectorSize(int N)
{
	if (N<8)
		return DskSectorSizeTable[N];
	else
		return 0x0200;
}

#include "ams.h"

void	FormatDsk(int FormatID, char *DiskImageName)
{
    const FORMAT_INFO *pFormatsRecognised = GetFormatsRecognised();
	const XDPB *Xdpb = pFormatsRecognised[FormatID].pFormatXdpb;
	int	NoOfTracks;
	int	NoOfSides = Xdpb->ReservedSelectFormat+1;
	int NoOfSectorsPerTrack = Xdpb->SectorsPerTrack;
	int	SectorSize = DskGetSectorSize(Xdpb->SectorSizeN);
	int TrackSize = (SectorSize*NoOfSectorsPerTrack)+sizeof(DSKTRACKHEADER);

	FILE	*fh;
	char	Header[256];
	int		i,j,k;
	char	*pSectorData;

	int		TotalBlocks,RecordsPerDisc,TotalSectors;

	/* calculate number of tracks */
	TotalBlocks = (Xdpb->TotalBlocksLow & 0x0ff) |
					((Xdpb->TotalBlocksHigh & 0x0ff)<<8);

	RecordsPerDisc = TotalBlocks * (Xdpb->RecordsPerBlock+1);

	TotalSectors = RecordsPerDisc/Xdpb->RecordsPerSector;

	NoOfTracks = (TotalSectors/(NoOfSectorsPerTrack*NoOfSides))+1;

	/* allocate memory for sector data */
	pSectorData = (char *)malloc(SectorSize);

	if (pSectorData!=NULL)
	{
		/* fill sector buffer with filler byte */
		memset(pSectorData,FILLER_BYTE,SectorSize);
	}

	/* open disk image file */
	fh = fopen(DiskImageName,"wb");

	if (fh!=NULL)
	{
		DSKHEADER	*pMainHeader = (DSKHEADER *)Header;

		/* initialise main header */

		/* clear main header */
		memset(pMainHeader,0,sizeof(DSKHEADER));

		/* copy main header id */
		memcpy(pMainHeader,HeaderText,sizeof(HeaderText));

		/* initialise header */
		pMainHeader->NumTracks = NoOfTracks;
		pMainHeader->NumSides = NoOfSides;
		pMainHeader->TrackSizeLow = (TrackSize & 0x0ff);
		pMainHeader->TrackSizeHigh = (TrackSize>>8);

		/* write main header */
		fwrite(pMainHeader,sizeof(DSKHEADER),1,fh);

		for (i=0; i<NoOfTracks; i++)
		{
			for (j=0; j<NoOfSides; j++)
			{
				/* initialise track header */

				DSKTRACKHEADER *pTrackHeader = (DSKTRACKHEADER *)Header;
				DSKCHRN			*pCHRN;

				/* clear track header */
				memset(pTrackHeader,0,sizeof(DSKTRACKHEADER));

				/* copy track header id */
				memcpy(pTrackHeader,TrackText,sizeof(TrackText));

				pTrackHeader->track = i;
				pTrackHeader->side = j;
				pTrackHeader->SPT = NoOfSectorsPerTrack;
				pTrackHeader->BPS = Xdpb->SectorSizeN;
				pTrackHeader->Gap3 = 0x052; //Xdpb->GapLengthFormat;
				pTrackHeader->FillerByte = FILLER_BYTE;

				/* write sector ids */
				pCHRN = &pTrackHeader->SectorIDs[0];

				for (k=0; k<NoOfSectorsPerTrack; k++)
				{
					pCHRN->C = i;
					pCHRN->H = j;
					pCHRN->R = Xdpb->FirstSectorNum + k;
					pCHRN->N = Xdpb->SectorSizeN;
					pCHRN++;
				}

				/* write track header */
				fwrite(pTrackHeader,sizeof(DSKTRACKHEADER),1,fh);

				if (pSectorData!=NULL)
				{
					/* write sector data */
					for (k=0; k<NoOfSectorsPerTrack; k++)
					{
						fwrite(pSectorData,SectorSize,1,fh);
					}
				}

			}
		}

		/* close output file */
		fclose(fh);
	}

	if (pSectorData!=NULL)
		free(pSectorData);
}

