/*****************************************************************************/
/* This module handles reading and writing files using standard C functions  */
/* This module is used for PC, Amiga, Mac etc                                */
/*****************************************************************************/

// CODE FOR PC (WINDOWS)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "host.h"
#include "namelist.h"

#ifdef _WIN32
#include	<io.h>
#endif

#ifdef UNIX
#include <sys/io.h>
#include <dirent.h>
#endif

#include "wildcard.h"
#include "filen.h"


void	OpenFileHost(CPC_FILEHANDLE *pCPCFileHandle, char *Filename, int Access)
{
char    AccessString[3];
int		NameLen;
//char	HostFilename[16];

	NameLen = strlen(Filename)+1;

	if (pCPCFileHandle!=NULL)
	{
		if (Access == CPC_FILEHANDLE_READ)
		{
			strncpy(AccessString,"rb",3);
		}

		if (Access == CPC_FILEHANDLE_WRITE)
		{
       		strncpy(AccessString,"wb",3);
		}

		pCPCFileHandle->fh = fopen(Filename,AccessString);

		printf("OpenFileHost: Executed.\n");
		printf("Filename: %s\n",pCPCFileHandle->Filename);
	}

}


int	GetCharHost(CPC_FILEHANDLE *pCPCFileHandle)
{
		return getc(pCPCFileHandle->fh);
}

void	PutCharHost(CPC_FILEHANDLE *pCPCFileHandle,int ch)
{
		putc(ch,pCPCFileHandle->fh);
}

void	CloseFileHost(CPC_FILEHANDLE *pCPCFileHandle)
{
	if (pCPCFileHandle!=NULL)
	{
		/* close file */
		if (pCPCFileHandle->fh!=NULL)
			fclose(pCPCFileHandle->fh);

		/* free space allocated to name */
		if (pCPCFileHandle->Filename!=NULL)
			free(pCPCFileHandle->Filename);
	}
}

int	GetFileSizeHost(CPC_FILEHANDLE *pCPCFileHandle)
{
#ifdef WIN32
	int	fno;
	fno = _fileno(pCPCFileHandle->fh);
	return	_filelength(fno);
#endif
#ifdef UNIX
	int pos = ftell(pCPCFileHandle->fh);
	fseek(pCPCFileHandle->fh,0, SEEK_END);
	int len = ftell(pCPCFileHandle->fh);
	fseek(pCPCFileHandle->fh, pos, SEEK_SET);
	return len;
#endif
}

BOOL	Host_FileFit(int DriveNum, int FileSize)
{
	return TRUE;
}


/* PC code to generate a list of files matching a filespec */
BOOL	GetFileList(char *FileSpec,void *pData, void (*pCallback)(void *,char *))
{
#ifdef UNIX
	return FALSE;
#else
	struct _finddata_t FindData;
	long	handle;

	/* find first filename matching file spec */
	handle = _findfirst("*.*",&FindData);

	/* error, no filenames matched filespec so return FALSE */
	if (handle==-1)
		return FALSE;

	if ((FindData.attrib & _A_SUBDIR)==0)
	{
		char	CPCFilename[16];

		/* convert to cpc filename to check */
		HostFilenameToCPCFilename(FindData.name, CPCFilename);

		if (WildCard_Match(CPCFilename, FileSpec))
		{
			/* add host name */
			NameList_AddName((NAME_LIST_HEADER *)pData, FindData.name);
		}
	}
	/* find next filename matching file spec */

	while (_findnext(handle, &FindData)==0)
	{
		if ((FindData.attrib & _A_SUBDIR)==0)
		{
			char	CPCFilename[16];

			/* convert to cpc filename to check */
			HostFilenameToCPCFilename(FindData.name, CPCFilename);

			if (WildCard_Match(CPCFilename, FileSpec))
			{
				/* add host name */
				NameList_AddName((NAME_LIST_HEADER *)pData, FindData.name);
			}

		}

	}

	/* finishing matching filenames */
	_findclose(handle);

	/* at least 1 file matched the filespec so return TRUE */
	return TRUE;
#endif
	}

/* callback to add a name to the list */
void	ListCallback(void *pData, char *Name)
{
	NAME_LIST_HEADER	*pList = (NAME_LIST_HEADER *)pData;

	NameList_AddName(pList,Name);
}

void	Host_GenerateFileList(char *FileSpec, NAME_LIST_HEADER *pNameList)
{
	GetFileList(FileSpec, pNameList, ListCallback);
}
