/* Code to extract/add files to disc images or real disk */
/* (c) Kevin Thacker */

#define DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "global.h"
#include "filexfer.h"
#include "diskimg.h"
#include "filen.h"
#include "host.h"
#include "dskfile.h"

static int     SectorSizeTable[7]={128,256,512,1024,2048,4096,8192};

static DRIVE_DATA       DriveData[2];

/* Reading information */
static FILELIST_NODE    *pFileList;

static int    InitParams(int DriveNum)
{
XDPB    *Xdpb;
DRIVE_DATA *pDriveData;

                pDriveData = &DriveData[DriveNum];

        Xdpb = pDriveData->Xdpb;

        pDriveData->XDPBSectorsPerBlock = ((Xdpb->RecordsPerBlock+1)/(Xdpb->RecordsPerSector));
        pDriveData->XDPBNumDirEntries = (Xdpb->NumDirEntriesHigh<<8)|(Xdpb->NumDirEntriesLow);
        pDriveData->XDPBSectorSize = GetSectorSize(Xdpb->SectorSizeN);
        pDriveData->XDPBBlockSize = (Xdpb->RecordsPerBlock+1)<<7;
        pDriveData->XDPBTotalBlocks = (Xdpb->TotalBlocksHigh<<8)|(Xdpb->TotalBlocksLow);
        pDriveData->XDPBNumDirBlocks = ((pDriveData->XDPBNumDirEntries+1)*sizeof(AMSDOSDIRENTRY))/pDriveData->XDPBBlockSize;

                  pDriveData->BlockBitmapSize = pDriveData->XDPBTotalBlocks+1;

                  if ((pDriveData->pBlockBitmap = malloc(pDriveData->BlockBitmapSize))==NULL)
                  {
                                return FALSE;
                  }

        pDriveData->CPCDirectorySize = pDriveData->XDPBNumDirBlocks*pDriveData->XDPBBlockSize;

                  if ((pDriveData->pCPCDirectory = malloc(pDriveData->CPCDirectorySize))==NULL)
                  {
                           return FALSE;
                  }

                  if ((pDriveData->pBlock = malloc(pDriveData->XDPBBlockSize))==NULL)
                  {
                                return FALSE;
                  }

                pDriveData->BlockListSize = (pDriveData->XDPBTotalBlocks+1)*sizeof(int);

                  if ((pDriveData->pBlockList = malloc(pDriveData->BlockListSize))==NULL)
                  {
                                return FALSE;
                  }


        printf("Directory size fix needed here!\n");

        printf("%d\n",pDriveData->XDPBNumDirBlocks*pDriveData->XDPBBlockSize);
        printf("%d\n",((pDriveData->XDPBNumDirEntries+1)*sizeof(AMSDOSDIRENTRY)));


        #ifdef DEBUG
        printf("XDPBSectorsPerBlock: %d\n",pDriveData->XDPBSectorsPerBlock);
        printf("XDPBNumDirEntries: %d\n",pDriveData->XDPBNumDirEntries);
        printf("XDPBSectorSize: %d\n",pDriveData->XDPBSectorSize);
        printf("XDPBBlockSize: %d\n",pDriveData->XDPBBlockSize);
        printf("AmsdosDirEntrySize: %d\n",sizeof(AMSDOSDIRENTRY));
        printf("AmsdosDiskHeaderSize: %d\n",sizeof(AMSDOSFILEHEADER));
        printf("CPCDirectorySize: %d\n",pDriveData->CPCDirectorySize);
        printf("BlockListSize: %d\n",pDriveData->BlockListSize);
        printf("XDPBTotalBlocks: %d\n",pDriveData->XDPBTotalBlocks);
        printf("BlockBitmapSize: %d\n",pDriveData->BlockBitmapSize);
        #endif

        return TRUE;
}

static void             FileXfer_FreeDrive(int DriveNum)
{
DRIVE_DATA *pDriveData;

                pDriveData = &DriveData[DriveNum];

           switch (pDriveData->DriveType)
           {
                case DISK_IMAGE:
                {
                        CloseDiskImage(pDriveData);
                        break;
                }
                case REAL_DISK:
                {
                        /*   */
                        break;
           }
        }
                if (pDriveData->pBlockBitmap!=NULL)
                {
                        free(pDriveData->pBlockBitmap);
                }

                if (pDriveData->pCPCDirectory!=NULL)
                {
                        free(pDriveData->pCPCDirectory);
                }

                if (pDriveData->pBlock!=NULL)
                {
                        free(pDriveData->pBlock);
                }

                if (pDriveData->pBlockList!=NULL)
                {
                        free(pDriveData->pBlockList);
                }
}


/* Read a sector of data from disk image or real disk,
multiple disk image formats can be supported here */


static int     ReadSector(SECTOR_POS *pSectorPos, char *pData)
{
        #ifdef DEBUG
        printf("Reading sector: Drive %d Track %d Sector %x Side %d\n",pSectorPos->Drive,pSectorPos->Track,pSectorPos->Sector,pSectorPos->Side);
        #endif

        switch (DriveData[pSectorPos->Drive].DriveType)
        {
            case DISK_IMAGE:
            {

                                return GetSectorDsk(&DriveData[pSectorPos->Drive],pSectorPos,pData);
                                        break;
                                }
                                case REAL_DISK:
                                {
/*                                      return GetSectorDisk(pSectorPos->Track,pSectorPos->Sector,pSectorPos->Side,pData);*/
                                        break;
                                }
                                case HOST_FILESYSTEM:
                                        return FALSE;
                        }
                return TRUE;

}

static XDPB    *GetFormat(int DriveNum)
{
FORMAT_DATA FormatData;
int     i;

                switch (DriveData[DriveNum].DriveType)
                {
                        case    DISK_IMAGE:
                        {
                        GetFormatDsk(&DriveData[DriveNum],&FormatData);
                                break;
                        }
                        case    REAL_DISK:
                        {
                                /*GetFormatDisk(&DriveData[DriveNum],&FormatData);*/
                                break;
                        }
                }

        for (i=0; i<NUM_RECOGNISED_FORMATS; i++)
        {
                if (memcmp(FormatsRecognised[i].FormatData,&FormatData,sizeof(FORMAT_DATA))==0)
                {
                        #ifdef DEBUG
                        printf("Format found: %s\n",FormatsRecognised[i].FormatName);
                        #endif
                        return FormatsRecognised[i].FormatXdpb;
                }
        }

        #ifdef DEBUG
        printf("Not recognised format\n");
        #endif
        return 0;
}

/* Read a sector of data from disk image or real disk,
multiple disk image formats can be supported here */

static int     WriteSector(SECTOR_POS *pSectorPos, char *pData)
{
        printf("Writing sector: Drive %d Track %d Sector %x Side %d",pSectorPos->Drive,pSectorPos->Track,pSectorPos->Sector,pSectorPos->Side);

        switch (DriveData[pSectorPos->Drive].DriveType)
        {
            case DISK_IMAGE:
            {

                                return PutSectorDsk(&DriveData[pSectorPos->Drive],pSectorPos,pData);
                                        break;
                                }
                                case REAL_DISK:
                                {
/*                                      return PutSectorDisk(pSectorPos->Track,pSectorPos->Sector,pSectorPos->Side,pData);*/
                                        break;
                                }
                                case HOST_FILESYSTEM:
                                        return FALSE;
                        }
        return TRUE;
}

/* Convert block number to sector offset which describes the number of
sectors to seek in to find start of block */

static int     BlockToSectorOffset(int DriveNum,int BlockNum)
{
        return (BlockNum*DriveData[DriveNum].XDPBSectorsPerBlock);
}

/* Convert sector offset to track,sector and side describing location
   of sector on disc */

static void    SectorOffsetToSectorPos(int DriveNum,int SectorOffset,SECTOR_POS *pSectorPos)
{
XDPB *Xdpb;

        Xdpb = DriveData[DriveNum].Xdpb;

        if (Xdpb->ReservedSelectFormat!=1)
        {
                /* Normal disc format */

                                        pSectorPos->Drive = DriveNum;
                pSectorPos->Sector=SectorOffset-((SectorOffset/Xdpb->SectorsPerTrack)*Xdpb->SectorsPerTrack)+Xdpb->FirstSectorNum;
                pSectorPos->Track=(SectorOffset/Xdpb->SectorsPerTrack)+((Xdpb->DirTrackHigh<<8)|(Xdpb->DirTrackLow));
                pSectorPos->Side=0;                     /* Temp */
        }
        else
        {
                        /* Double sided disc formats */

                        pSectorPos->Drive = DriveNum;
         pSectorPos->Sector = SectorOffset-((SectorOffset/Xdpb->SectorsPerTrack)*Xdpb->SectorsPerTrack)+Xdpb->FirstSectorNum;
         pSectorPos->Track = (SectorOffset/(Xdpb->SectorsPerTrack*2))+((Xdpb->DirTrackHigh<<8)||(Xdpb->DirTrackLow));

         if (SectorOffset-(pSectorPos->Track*(Xdpb->SectorsPerTrack*2))>=(Xdpb->SectorsPerTrack))
         {
           pSectorPos->Side = 1;
         }
         else
         {
           pSectorPos->Side = 0;
         }
        }
}

/* Read a block from disc and put into buffer */

static int     ReadBlock(int DriveNum,int BlockNumber,char *pData)
{
int i;
int SectorOffset;
SECTOR_POS SectorPos;
char *pDataPos;

        #ifdef DEBUG
        printf("Read Block: %d\n",BlockNumber);
        #endif

        pDataPos = pData;
        SectorOffset=BlockToSectorOffset(DriveNum,BlockNumber);

        for (i=0; i<DriveData[DriveNum].XDPBSectorsPerBlock; i++)
        {
                SectorOffsetToSectorPos(DriveNum,SectorOffset,&SectorPos);
                if (ReadSector(&SectorPos,pDataPos)==FALSE)
                        return FALSE;
                pDataPos+=DriveData[DriveNum].XDPBSectorSize;
                SectorOffset++;
        }
        return TRUE;
}

/* Write a block to disc from buffer */

static int     WriteBlock(int DriveNum,int BlockNumber,char *pData)
{
int i;
int SectorOffset;
SECTOR_POS SectorPos;
char    *pDataPos;

        pDataPos = pData;
        SectorOffset = BlockToSectorOffset(DriveNum,BlockNumber);

        for (i=0; i<DriveData[DriveNum].XDPBSectorsPerBlock; i++)
        {
                SectorOffsetToSectorPos(DriveNum,SectorOffset,&SectorPos);
                if (WriteSector(&SectorPos,pDataPos)==FALSE)
                        return FALSE;
                pDataPos+=DriveData[DriveNum].XDPBSectorSize;
                SectorOffset++;
        }
        return TRUE;
}

/* Read Directory into buffer */

static int     ReadDirectory(int DriveNum)
{
int     i;
char    *pData;

        pData=(char *)DriveData[DriveNum].pCPCDirectory;

        for (i=0; i<DriveData[DriveNum].XDPBNumDirBlocks; i++)
        {
                if (ReadBlock(DriveNum,i,pData)==FALSE)
                        return FALSE;
                pData+=DriveData[DriveNum].XDPBBlockSize;
        }
        return TRUE;
}


/* Write Directory from buffer */

static int     WriteDirectory(int DriveNum)
{
int     i;
char    *pData;



        pData=(char *)DriveData[DriveNum].pCPCDirectory;

        for (i=0; i<DriveData[DriveNum].XDPBNumDirBlocks; i++)
        {
                if (WriteBlock(DriveNum,i,pData)==FALSE)
                        return FALSE;
                pData+=DriveData[DriveNum].XDPBBlockSize;
        }
        return TRUE;
}

/* Find free dir entry to add data to, NULL=DirectoryFull */

static AMSDOSDIRENTRY *GetFreeDirEntry(int DriveNum)
{
int     i;
AMSDOSDIRENTRY  *pData;

        pData=(AMSDOSDIRENTRY *)DriveData[DriveNum].pCPCDirectory;

        for (i=0; i<DriveData[DriveNum].XDPBNumDirEntries+1; i++)
        {
                /* User of 0x0e5 indicates deleted entry or free entry */

                if (pData->User==0x0e5)
                        return pData;
                pData++;
        }

        return NULL;
}

static void    GetAmsFilename(char *Filename,AMSDOSDIRENTRY *pCPCDirectoryEntry)
{
int i;
        /* copy filename */

        memcpy(&Filename[0],pCPCDirectoryEntry->Filename,8);
        Filename[8]='.';
        memcpy(&Filename[9],pCPCDirectoryEntry->Extension,3);

        /* remove any bits which say about r/w etc */

        for (i=0; i<3; i++)
        {
                Filename[9+i] = (char)(Filename[9+i] & 0x07f);
        }
        Filename[12] = '\0';
}

static void		PutAmsFilename(char *Filename, AMSDOSDIRENTRY *pCPCDirectoryEntry)
{

		memcpy(pCPCDirectoryEntry->Filename,&Filename[0],8);
		memcpy(pCPCDirectoryEntry->Extension,&Filename[9],3);
}


static void     UpdateBlockBitmap(int DriveNum)
{
int i,j;
int Block;
AMSDOSDIRENTRY *pCPCDirectory;
char *pBlockBitmap;
XDPB *Xdpb;

        memset(DriveData[DriveNum].pBlockBitmap,0x000,DriveData[DriveNum].BlockBitmapSize);

        pCPCDirectory = (AMSDOSDIRENTRY *)DriveData[DriveNum].pCPCDirectory;
        pBlockBitmap = (char *)DriveData[DriveNum].pBlockBitmap;
        Xdpb = DriveData[DriveNum].Xdpb;

        for (i=0; i<DriveData[DriveNum].XDPBNumDirEntries; i++)
        {
                if (pCPCDirectory[i].User!=0x0e5)
                {
                        /* Directory entry used */

                        if (Xdpb->ReservedSelectFormat==0)
                        {
                                for (j=0; j<16; j++)
                                {
               Block = pCPCDirectory[i].Blocks[j] & 0x0ff;
                                        pBlockBitmap[Block] = 0x0ff;

                                }
                        }
                        else
                        {
                                for (j=0; j<8; j++)
                                {
               Block = (pCPCDirectory[i].Blocks[(j*2)+1] & 0x0ff)|(pCPCDirectory[i].Blocks[j*2] & 0x0ff);
                                        pBlockBitmap[Block] = 0x0ff;
                                }
                        }
                }
        }
}

static void    MakeBlockListRead(int DriveNum,char *Filename,int UserNo)
{
int     DirEntryBlock;
int     i,j;
char    AmstradFilename[16];
AMSDOSDIRENTRY *pCPCDirectory;
int     *pBlockList;
XDPB    *Xdpb;

        pCPCDirectory = (AMSDOSDIRENTRY *)DriveData[DriveNum].pCPCDirectory;
        pBlockList = DriveData[DriveNum].pBlockList;
        Xdpb = DriveData[DriveNum].Xdpb;

        /* Clear Block List */

        memset(DriveData[DriveNum].pBlockList,0,DriveData[DriveNum].BlockListSize);

        /* Get all blocks this file uses */

        for (i=0; i<DriveData[DriveNum].XDPBNumDirEntries+1; i++)
        {
          if ((pCPCDirectory[i].User!=0x0e5) && (pCPCDirectory[i].User == UserNo))
          {
                GetAmsFilename(AmstradFilename,&pCPCDirectory[i]);

                if (strcmp(AmstradFilename,Filename)==0)
                {
                  /* Filename matches */

                  /* Copy blocks into our list at correct place */

                if (Xdpb->ReservedSelectFormat==0)
                {
                  for (j=0; j<16; j++)
                  {

                        DirEntryBlock = (pCPCDirectory[i].DirEntryNum & 0x0ff);
                        pBlockList[(DirEntryBlock<<4)+j] = pCPCDirectory[i].Blocks[j] & 0x0ff;
                  }
                }
                else
                {
                  for (j=0; j<8; j++)
                  {
                        DirEntryBlock = (pCPCDirectory[i].DirEntryNum & 0x0ff);
                        pBlockList[(DirEntryBlock<<3)+j] = (pCPCDirectory[i].Blocks[(j*2)+1] & 0x0ff)|(pCPCDirectory[i].Blocks[j*2] & 0x0ff);
                  }
                }
                }
          }

        }
}

static void    MakeFreeBlockList(int DriveNum)
{
int     i,j;
int     *pBlockList;
char	*pBlockBitmap;

        pBlockList = DriveData[DriveNum].pBlockList;
		  pBlockBitmap = DriveData[DriveNum].pBlockBitmap;

		j=0;

		for (i=0; i<DriveData[DriveNum].XDPBTotalBlocks; i++)
		{
			/* Is block used ? */
			if (pBlockBitmap[i]==0x00)
			{
				/* No, store block number in our list */
				pBlockList[j] = i;
				j++;
			}
	   }
		/* Finish List */

		pBlockList[j] = 0;
}




/* Return sector size from N value in sector ID */

static int     GetSectorSize(int N)
{
        if (N<8)
                return SectorSizeTable[N];
        else
                return 0x0200;
}

/*
static void    StringToAmsFilename(char *String,char *Filename)
{
char    ch;
int     index,index2;
int i;

        memset(&Filename[0],' ',12);
        Filename[8]='.';
        Filename[12]='\0';

        index = 0;
        index2 = 0;

        ch='\0';

        do
        {
                ch = String[index];
                index++;

                if ((ch!=' ') && (ch!='.') && (ch!='\0'))
                {
                        Filename[index2] = ch;
                        index2++;
                }
        }
        while ((ch!=' ') && (ch!='.') && (ch!='\0'));

        if (ch=='\0')
        {
                goto makecaps;
        }


        if (ch==' ')
        {
                do
                {
                        ch = String[index];
                        index++;
                        if ((ch!=' ') || (ch=='\0'))
                        {
                                break;
                        }
                }
                while ((ch==' ') && (ch!='\0'));

                if (ch=='\0')
                {
                        goto makecaps;
                }
        }

        index2 = 9;

        do
        {
                ch = String[index];
                index++;

                if ((ch!=' ') && (ch!='\0'))
                {
                        Filename[index2] = ch;
                        index2++;
                }
        }
        while ((ch!=' ') && (ch!='\0'));

makecaps:
        for (i=0; i<12; i++)
        {
                Filename[i] = toupper(Filename[i]);
        }

        printf(Filename);

}
*/

/* Opens a file for reading or writing on a specific drive */

static CPC_FILEHANDLE *OpenFile(int DriveNum,char *Filename, int Access)
{
        switch (DriveData[DriveNum].DriveType)
        {
            case DISK_IMAGE:
            {

                                return OpenFileDsk(DriveNum,Filename,Access);
                                        break;
                                }
                                case REAL_DISK:
                                {
/*                                      return OpenFileDisk(DriveNum,Filename,Access);
                                        break;
                                }
                                case HOST_FILESYSTEM:
                                        return OpenFileHost(DriveNum,Filename,Access);
                        }




}

static int		GetChar(CPC_FILEHANDLE *pCPCFileHandle)
{
        switch (DriveData[pCPCFileHandle->Drive].DriveType)
        {
            case DISK_IMAGE:
            {

                                return GetCharDsk(pCPCFileHandle);
                                        break;
                                }
                                case REAL_DISK:
                                {
/*                                      return PutCharDisk(pCPCFileHandle);*/
                                        break;
                                }
                                case HOST_FILESYSTEM:
											{
													return PutCharHost(pCPCFileHandle);
													break;
			                        }

}

static void		PutChar(CPC_FILEHANDLE *pCPCFileHandle,int Ch)
{
        switch (DriveData[pCPCFileHandle->Drive].DriveType)
        {
            case DISK_IMAGE:
            {

                                return PutCharDsk(pCPCFileHandle,Ch);
                                        break;
                                }
                                case REAL_DISK:
                                {
/*                                      return PutCharDisk(pCPCFileHandle,Ch);*/
                                        break;
                                }
                                case HOST_FILESYSTEM:
                                        return PutCharHost(pCPCFileHandle,Ch);
                        }



}

/* Closes a file for reading or writing on a specific drive */

static void		CloseFile(CPC_FILEHANDLE *pCPCFileHandle)
{
        switch (DriveData[pCPCFileHandle->Drive].DriveType)
        {
            case DISK_IMAGE:
            {

               return CloseFileDsk(pCPCFileHandle);
                                        break;
                                }
                                case REAL_DISK:
                                {
/*                                      return CloseFileDisk(pCPCFileHandle);*/
                                        break;
                                }
                                case HOST_FILESYSTEM:
													return CloseFileHost(pCPCFileHandle);
                        }


}

static void     InitialiseDrive(int DriveNum,int DriveType,void *pData)
{
        DriveData[DriveNum].DriveType = DriveType;

        switch (DriveType)
        {
                case    REAL_DISK:
                {
                        printf("Real Disk selected.\nReal Disk not supported!\n");
                        break;
                }
                case    DISK_IMAGE:
                {
                        printf("Disk Image selected\n");
                        OpenDiskImage(&DriveData[DriveNum],(char *)pData);
                        break;
                }
                case    HOST_FILESYSTEM:
                {
                        printf("Host filesystem selected\n");
                        break;
                }
        }

        /* Identify disc format */

        DriveData[DriveNum].Xdpb = GetFormat(DriveNum);

        /* Setup parameters */

        InitParams(DriveNum);

                ReadDirectory(DriveNum);

                UpdateBlockBitmap(DriveNum);

                }

int main(int argc, char *argv[])
{

      InitialiseDrive(0,DISK_IMAGE,(void *)argv[1]);

/*      ReadFile(0,Filename,0);*/

      FreeDrive(0);

      return TRUE;
}

